/********************************************************************/
/* File: symtab.c                                                   */
/* A symbol table implemented as a chained hash table               */
/********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "symtab.h"
#include "token.h"

/* maximum size of hash table */
#define SIZE 200 
#define MAXTOKENLEN 40
#define MAXMEMSIZE 50 

/* power of two multiplier in hash function */
#define SHIFT 4


char * memory[MAXMEMSIZE]; // an array of words, each 256 bytes */ 
int mem_offset=0;  // stored in symbol table as an offset into pseudo-memory

/* the hash function */
static int hash ( char * key )
{ int temp = 0;
  int i = 0;
  while (key[i] != '\0')
  { temp = ((temp << SHIFT) + key[i]) % SIZE;
    ++i;
  }
  return temp;
}

/* a linked list of references (line nos) for each variable */
typedef struct RefListRec { 
     int lineno;
     struct RefListRec * next;
} * RefList;


/* hash entry holds variable name and its reference list */
typedef struct HashRec { 
     char st_name[MAXTOKENLEN+1];
     int st_size;
     Token* st_token;
     int st_value;
     RefList lines;
     struct HashRec * next;
} * Node;

/* the hash table is a static array of node pointers */
static Node hashTable[SIZE];

/*  insert entry - if entry exists just add reference line no.  */
void insert( char * name, int len, int lineno, Token* tok )
{ int h = hash(name);
  Node l =  hashTable[h];
  while ((l != NULL) && (strcmp(name,l->st_name) != 0))
     l = l->next;
  if (l == NULL) /* variable not yet in table */
  {   l = (Node) malloc(sizeof(struct HashRec));
      strncpy(l->st_name, name, len);  
      l->st_token = tok;
      /* grab a chunk of memory */
      mem_offset++;
      memory[mem_offset] = malloc(sizeof(char[256]));
      if (tok->data == 'R')
         strncpy(memory[mem_offset],name,len);

     l->st_value = mem_offset;

     l->lines = (RefList) malloc(sizeof(struct RefListRec));
     l->lines->lineno = lineno;
     l->lines->next = NULL;
     l->next = hashTable[h];
     hashTable[h] = l; 
  }
  else /* found in table, so just add line number */
  {  RefList t = l->lines;
     while (t->next != NULL) t = t->next;
     t->next = (RefList) malloc(sizeof(struct RefListRec));
     t->next->lineno = lineno;
     t->next->next = NULL;
  }
} 

/* return 1 if found or 0 if not found */
int lookup ( char * name )
{  int h = hash(name);
   Node l =  hashTable[h];
   while ((l != NULL) && (strcmp(name,l->st_name) != 0))
     l = l->next;
   if (l == NULL) return 0;
   else return (1);
}

Token* getToken ( char * name )
{  int h = hash(name);
   Node l =  hashTable[h];
   while ((l != NULL) && (strcmp(name,l->st_name) != 0))
     l = l->next;
   if (l == NULL) return 0;
   else return l->st_token;
}


/* print to stdout by default */ 
void dump_symtab(FILE * of)
{ int i;
  fprintf(of,"Name       Type       Data   Value Lines\n");
  fprintf(of,"---------- ---------- ------ ----- ---------------------\n");
  for (i=0;i<SIZE;++i)
  { if (hashTable[i] != NULL)
    { Node l = hashTable[i];
      while (l != NULL)
      { RefList t = l->lines;
        fprintf(of,"%-10s ",l->st_token->name);
	switch(l->st_token->type)
	{
	  case 'L':
	    fprintf(of, "Literal    ");
	    break;
	  case 'I':
	    fprintf(of, "Identifier ");
	    break;
	  case 'C':
	    fprintf(of, "Const. ID  ");
	    break;
	  case 'O':
	    fprintf(of, "Operator   ");
	    break;
	  case 'F':
	    fprintf(of, "Function   ");
	    break;
	  default:
	    fprintf(of, "Unknown    ");
	    break;
	}
	switch(l->st_token->data)
	{
	  case 'V':
	    fprintf(of, "Void   ");
	    break;
	  case 'R':
	    fprintf(of, "Real   ");
	    break;
	  case 'I':
	    fprintf(of, "Int    ");
	    break;
	  case 'C':
	    fprintf(of, "Char   ");
	    break;
	  case 'S':
	    fprintf(of, "String ");
	    break;
	  case 'F':
	    fprintf(of, "Func   ");
	    break;
	  default:
	    fprintf(of, "???    ");
	    break;
	}
        fprintf(of,"%-5d ",l->st_value);
        while (t != NULL)
        { fprintf(of,"%-5d ",t->lineno);
          t = t->next;
        }
        fprintf(of,"\n");
        l = l->next;
      }
    }
  }
}


void dump_mem () {
   int i;
   //for (i = 0; i <= mem_offset; i++)
   //   fprintf(stdout, "%s\n", memory[i]);
   /* cleanup */
  for (i = 0; i <= mem_offset; i++)
      free (memory[i]);
}

